#!/usr/bin/python

program_name = "pfm"
program_version = "0.1"


print "Loading GTK..."

from gtk import *

print "Loading GTK...  done."

def runqueue():
  while events_pending():
    mainiteration()


#
# Config file handling, and defaults...
#

######################################################################
# Config class:
#
def is_config(obj):
  return hasattr(obj,'is_config') and obj.is_config

class config:
  is_config = 1

  def __init__(self):
    import os
    self.options = {}
    self.usrfilename = "%s/.%s.conf" % (os.environ["HOME"], program_name)
    self.globalfilename = "/etc/%s/%s.conf" % (program_name, program_name)
    self.namespace = {
      "options" : self.options,
      }

  ##################################################
  #
  def load(self):
    import os.path
    for name in [self.globalfilename, self.usrfilename]:
      if os.path.exists(name):
        self.parse(name)

  ##################################################
  #
  def parse(self, filename):
    #print "parsing %s..." % filename
    
    try:
      execfile(filename, self.namespace)
    except:
      print "Error in config file %s" % filename
      import traceback, sys
      foo = traceback.format_exception(sys.exc_type,
                                       sys.exc_value,
                                       sys.exc_traceback)
      for item in foo:
        item = item[:-1]
        print item
      print "Press Enter to continue..."
      raw_input()

######################################################################
#
class file:

  def __init__(self):
    # basename
    # type: dir, file, executable
    # size
    # dates: mtime, ctime, atime
    # more?
    pass
      
######################################################################
#
class dir:

  def __init__(self, cfg, path="."):
    self.path = path
    self.cfg = cfg
    self.files = []
    self.selected = []

    self.typeorder = {
      " " : -10,
      "/" : 0,
      "f" : 10,
      }


  def loaddir(self, path=""):
    import os
    import os.path
    import stat
    
    if path:
      if path[0] == "/":
        self.path = path
      else:
        self.path = "%s/%s" % (self.path, path)

      self.path = os.path.abspath(self.path)
      
    self.files = []
    filenames = os.listdir(self.path)
    print filenames

    for name in filenames:
      f = file()
      f.basename = name
      fullpath = "%s/%s" % (self.path, name)
      s = os.lstat(fullpath)
      m = s[stat.ST_MODE]
      # TODO: support more file types
      if stat.S_ISDIR(m):
        f.type = "/"
      elif stat.S_ISREG(m):
        f.type = "f"
      else:
        f.type = " "

      f.size = s[stat.ST_SIZE]
      self.files.append(f)

    self.files.sort(self._sorter)

    for f in self.files:
      print "%s: %sb (%s)" % (f.basename, f.size, f.type)

  ##################################################
  #
  def _sorter(self, a, b):
    if a.type == b.type:
      if a.basename < b.basename: return -1
      else: return 1
    else:
      return self.typeorder[a.type] - self.typeorder[b.type]

  
######################################################################
#
class dirwin:

  def __init__(self, cfg, path="."):
    self.cfg = cfg
    self.dir = dir(cfg, path)

    self._updating_now = 0
    self._selected = []
    self._cursor = 0

    self.win = GtkWindow()
    self.win.connect("destroy", self.destroy)
    self.win.set_title("pfm") # TODO: add dir to title
    self.win.set_default_size(self.cfg.options["width"],
                              self.cfg.options["height"])

    self.vbox = GtkVBox()
    self.win.add(self.vbox)

    self.make_menus()

    self.sw = GtkScrolledWindow()
    self.sw.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)
    self.vbox.pack_start(self.sw, fill=TRUE, expand=TRUE)

    self.status = GtkLabel()
    self.vbox.add(self.status, fill=1, expand=0)

    self._key_actions = {
      GDK.Left   : self._up_dir,
      GDK.Right  : self._activate,
      GDK.Return : self._activate,
      GDK.Up     : self._up_arrow,
      GDK.Down   : self._down_arrow,
      }

    self.clist = GtkCList(1)
    self.clist.set_column_auto_resize(0,1)
    self.sw.add(self.clist)
    self.clist.connect("select-row", self._item_queue_cb)
    self.clist.connect("unselect-row", self._item_dequeue_cb)
    self.clist.connect("key-press-event", self._clist_keypress)
    #self.clist.set_selection_mode(SELECTION_BROWSE)
    self.clist.set_selection_mode(SELECTION_MULTIPLE)

  ##################################################
  #
  def make_menus(self):
    self.ag = GtkAccelGroup()
    self.win.add_accel_group(self.ag)
    self.itemf = GtkItemFactory(GtkMenuBar, "<main>", self.ag)
    # Keypresses for quick access to commands...
    ks = [
      ("/File",       None, None, 0, "<Branch>"),
      ("/File/quit",  "<Control>Q", self.destroy, 0, ""),
      ("/Edit",       None, None, 0, "<Branch>"),
      ("/Edit/(de)Select _All",  "<Control>A", self._select_all_cb, 0, ""),
      ("/Action",       None, None, 0, "<Branch>"),
      ]
    self.itemf.create_items(ks)
    self.menubar = self.itemf.get_widget("<main>")
    self.vbox.pack_start(self.menubar, fill=0, expand=0)


  ##################################################
  #
  def update_gui(self):
    #print "update_gui()"
    self.clist.clear()

    runqueue()

    self.clist.freeze()
    for f in self.dir.files:
      title = f.basename
      if f.type == "/":
        title = title + "/"
      elif f.type == "*":
        title = title + "/"
      data = [title]
      self.clist.append(data)
    self.clist.thaw()

  ##################################################
  #
  def go_dir(self, path="."):
    olddir = self.dir.path
    
    self.dir.loaddir(path)
    self._cursor = 0

    if path == "..":
      import os.path
      dirs = os.path.split(olddir)
      olddir = dirs[-1]
      #print "Old dir: %s" % (olddir)

      
    self.update()

    if path == "..":
      self.select_by_name(olddir)

  ##################################################
  # TODO: make it work so it can match based on just the first letter
  #       for fast selection of files...
  # 
  def select_by_name(self, path):
    self._cursor = 0
    return
  
    print "select_by_name(%s)" % (path)
    self._cursor = 0
    for i in xrange(len(self.dir.files)):
      compare = self.dir.files[i].basename
      print "%s <--> %s" % (path, compare)
      if path >= compare:
        self._cursor = i

    print "Setting cursor to... %s" % (self._cursor)
    self.clist.moveto(self._cursor, 0)
    self.clist.select_row(self._cursor, 0)

  ##################################################
  #
  def run(self):

    self.win.show_all()
    self.go_dir()
    mainloop()

  ##################################################
  #
  def update(self, *args):
    # Don't interrupt myself due to double-clicks
    if self._updating_now:
      return
    self._updating_now = 1  # cheapass mutex

    self.status.set_text("Hello.")

    self.update_gui()

    self._updating_now = 0

  ##################################################
  #
  def _select_all_cb(self, *args):
    if len(self._selected) > 0:
      self._unselect_all()
    else:
      self._select_all()

  ##################################################
  #
  def _select_all(self, *args):
    self.clist.select_all()

  ##################################################
  #
  def _unselect_all(self, *args):
    self.clist.unselect_all()

  ##################################################
  #
  def _item_select_cb(self, clist, row, col, gdkevent):
    #self.cursor = self.dir[row]
    foo = self.dir.files[row]
    print "Current: %s" % (foo.basename)

  ##################################################
  #
  def _item_queue_cb(self, clist, row, col, gdkevent):
    #self.cursor = self.dir[row]
    self._cursor = row
    foo = self.dir.files[row]
    self._selected.append(foo)
    print "Select: %s" % (foo.basename)

  ##################################################
  #
  def _item_dequeue_cb(self, clist, row, col, gdkevent):
    #self.cursor = self.dir[row]
    self._cursor = row
    foo = self.dir.files[row]
    try:
      self._selected.remove(foo)
    except ValueError:
      pass
    print "Deselect: %s" % (foo.basename)

  ##################################################
  #
  def _clist_keypress(self, cl, event, *args):
    if self._key_actions.has_key(event.keyval):
      self._key_actions[event.keyval]()
    else:
      print "Keypress: %s" % (event)

  ##################################################
  #
  def _up_arrow(self, *args):
    if self._cursor > 0:
      self._cursor = self._cursor - 1
    print "Cursor: %s" % (self._cursor)

  ##################################################
  #
  def _down_arrow(self, *args):
    if self._cursor < len(self.dir.files) - 1:
      self._cursor = self._cursor + 1
    print "Cursor: %s" % (self._cursor)

  ##################################################
  #
  def _up_dir(self, *args):
    print "Up dir!"
    self.go_dir("..")

  ##################################################
  #
  def get_selected_files(self):
    print "get_selected_files()"
    val = []
    current = self.dir.files[self._cursor]
    if len(self._selected) < 1:
      # return item under the cursor
      val = [current]
    else:
      #return all selected items, plus item under the cursor
      for f in self._selected:
        print f.basename
      if current in self._selected:
        val.extend(self._selected)
      else:
        val.append(current)
        val.extend(self._selected)

    return val

  ##################################################
  #
  def _activate(self, *args):
    print "Activate!"
    filestring = ""
    myfiles = self.get_selected_files()

    if myfiles[0].type == "/":
      self.go_dir(myfiles[0].basename)
      return
    
    for f in myfiles:
      filestring = "%s'%s' " % (filestring, f.basename)
    print "command %s" % (filestring)

  ##################################################
  #
  def destroy(self, *args):
    self.win.hide()
    self.win.unrealize()
    mainquit()
    del self.win

######################################################################
#
def run(args):
  cfg = config()
  cfg.load()
  
  p = dirwin(cfg)
  p.run()


######################################################################
#
if __name__ == "__main__":
  import sys
  run(sys.argv[1:])
