#!/usr/bin/python

program_name = "pfm"
program_version = "0.3"

#
# TODO:
# * convert main list to a TreeView instead of depreciated CList
# * gtk1 couldn't do...
#   * read cursor position!
#   * set cursor to prev directory when going to ".."!
#   * set cursor to prev position when reloading, updating, etc
#   * jump to file by first letter (can't set cursor position..)
#
# * auto-resize window
# - truncate label at bottom so it doesn't make the window resize?
# - make delete/rename windows fit better on small screens
#   (buttons overlap, entry goes past end of window, etc)
# - add key shortcuts to all dialog window options
# - option for "deselect all after launching"
#
# + delete file(s)
#   * with confirmation
#   - just do it
# * Rename file(s)
# - Copy/move file(s)
#   - Use x11 clipboard to send filenames between pfm windows?
#
# - prompt for, and automatically save actions?
#


num_dirs = 0

import pygtk
pygtk.require("2.0")
import gtk

from tktk import ButtonChoiceWindow, EntryWindow, Label, Window, runqueue

def num_unit(num):
  bytes = num
  unit = ""
  if bytes > 9999:
    unit = "K"
    bytes = bytes / 1024
  if bytes > 9999:
    unit = "M"
    bytes = bytes / 1024
  if bytes > 9999:
    unit = "G"
    bytes = bytes / 1024
  if bytes > 9999:
    unit = "T"
    bytes = bytes / 1024

  return "%s%s" % (bytes, unit)

class file:
  def __init__(self):
    # basename
    # type: dir, file, executable
    # size
    # dates: mtime, ctime, atime
    # more?
    pass
      
class dir:
  def __init__(self, cfg, path="."):
    self.path = path
    self.cfg = cfg
    self.files = []
    self.selected = []
    self.sorttype = "normal"

    self.typeorder = {
      " " : -10,
      "/" : 0,
      "f" : 10,
      }


  def loaddir(self, path=""):
    import os
    import os.path
    import stat
    
    if path:
      if path[0] == "/":
        self.path = path
      else:
        self.path = "%s/%s" % (self.path, path)


      self.path = os.path.abspath(self.path)
      
    self.files = []
    filenames = os.listdir(self.path)
    if filenames == []:
      filenames = [".."]
    #print filenames

    for name in filenames:
      f = file()
      f.basename = name
      fullpath = "%s/%s" % (self.path, name)
      try:
        s = os.stat(fullpath)
        m = s[stat.ST_MODE]
        # TODO: support more file types
        if stat.S_ISDIR(m):
          f.type = "/"
        elif stat.S_ISREG(m):
          f.type = "f"
        else:
          f.type = " "

        f.size = s[stat.ST_SIZE]
      except:
        f.size = 0
        f.type = " "

      self.files.append(f)

    if self.sorttype == "normal":
      self.files.sort(self._sorter)
    elif self.sorttype == "random":
      self.files.sort(self._rand_sorter)

    #for f in self.files:
    #  print "%s: %sb (%s)" % (f.basename, f.size, f.type)

  def _sorter(self, a, b):
    if a.type == b.type:
      if a.basename < b.basename: return -1
      else: return 1
    else:
      return self.typeorder[a.type] - self.typeorder[b.type]

  def _rand_sorter(self, a, b):
    if a.type == b.type:
      import whrandom
      return whrandom.randint(-100, 100)
    else:
      return self.typeorder[a.type] - self.typeorder[b.type]

class dirwin:
  def __init__(self, cfg, path="."):
    self.cfg = cfg
    self.dir = dir(cfg, path)

    self._updating_now = 0
    self._selected = []
    self._cursor = 0
    self._isearch = ""
    self._last_was_isearch = 0
    self._allow_select = [0]

    self.win = gtk.Window()
    self.win.connect("destroy", self.destroy)
    self.win.set_title("pfm") # TODO: add dir to title
    self.win.set_default_size(self.cfg.width,
                              self.cfg.height)

    self.vbox = gtk.VBox()
    self.win.add(self.vbox)

    self.make_menus()

    self.sw = gtk.ScrolledWindow()
    # POLICY_AUTOMATIC seems to scroll to the right on its own, which
    # is really annoying.  Better to make the window resize itself.
    self.sw.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
    self.vbox.pack_start(self.sw, 1, 1)

    #self.status_sw = gtk.ScrolledWindow()
    #self.status_sw.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
    self.status = gtk.Label()
    self.status.set_justify(gtk.JUSTIFY_LEFT)
    a = gtk.Alignment(0.0, 0.5, 0.0, 0.0)
    a.add(self.status)
    self.vbox.pack_start(a, 0, 0)
    #self.status_sw.add_with_viewport(a)
    #self.vbox.pack_start(self.status_sw, 0, 0)

    self._shift_actions = {
      #gtk.gdk.keyval_from_name("Up")     : self._key_up,
      #gtk.gdk.keyval_from_name("Down")   : self._key_down,
      gtk.gdk.keyval_from_name("Return") : self._nop, # handled by menu
      gtk.gdk.keyval_from_name("Right")  : self._activate_other,
      }

    self._key_actions = {
      gtk.gdk.keyval_from_name("Left")   : self._up_dir,
      gtk.gdk.keyval_from_name("Right")  : self._activate,
      gtk.gdk.keyval_from_name("Return") : self._nop, # handled by menu
      #gtk.gdk.keyval_from_name("space")  : self._file_info,
      #gtk.gdk.keyval_from_name("Up")     : self._key_up,
      #gtk.gdk.keyval_from_name("Down")   : self._key_down,
      #gtk.gdk.keyval_from_name("Next")   : self._key_next,
      gtk.gdk.keyval_from_name("BackSpace")  : self._backspace,
      #gtk.gdk.keyval_from_name("Escape")  : self._nop,
      }

    self._make_list()

  def _make_list(self):
    self.list = gtk.ListStore(type("foo"))
    self.listv = gtk.TreeView(self.list)
    render = gtk.CellRendererText()
    col = gtk.TreeViewColumn("file", render, text=0)
    self.listv.append_column(col)
    self.listv.set_headers_visible(0)
    #a = gtk.Alignment(1.0, 0.5, 1.0, 1.0)
    #a.add(self.listv)
    #self.sw.add(a)
    self.sw.add(self.listv)

    self.listv.connect("key-press-event", self._listcb_keypress)
    #self.listv.connect("move-cursor", self._show_args, "move-cursor")
    self.listv.connect("cursor-changed", self._listcb_move_cursor)
    #self.listv.connect("cursor-changed", self._show_args, "cursor-changed")
    #self.listv.connect("toggle-cursor-row", self._show_args, "toggle-cursor-row")
    #self.listv.connect("row-activated", self._show_args, "row-activated")
    self.listv.connect("row-activated", self._toggle_row)
    #self.listv.connect("select-cursor-row", self._show_args, "select-cursor-row")
    #self.listv.connect("toggle-cursor-row", self._show_args, "toggle-cursor-row")
    #self.listv.connect("select-all", self._show_args, "select-all")
    #self.listv.connect("unselect-all", self._show_args, "unselect-all")

    self.selection = self.listv.get_selection()
    #self.selection.set_mode(gtk.SELECTION_BROWSE)
    self.selection.set_mode(gtk.SELECTION_MULTIPLE)
    #self.selection.connect("changed", self._selection_changed)
    self.selection.set_select_function(self._selection_toggle)

  def make_menus(self):
    self.ag = gtk.AccelGroup()
    self.win.add_accel_group(self.ag)
    self.itemf = gtk.ItemFactory(gtk.MenuBar, "<main>", self.ag)

    ltext = "Launch"
    itext = "Choose..."
    if self.cfg.activate_interactive:
      ltext = "Choose..."
      itext = "Launch"

    # Keypresses for quick access to commands...
    ks = [
      ("/File",       None, None, 0, "<Branch>"),
      ("/File/Re_load",  "<Control>L", self._reload, 0, ""),
      ("/File/_New View",  "<Control>N", self._spawn_view, 0, ""),
      ("/File/_Quit",  "<Control>Q", self.destroy, 0, ""),
      ("/Edit",       None, None, 0, "<Branch>"),
      ("/Edit/(de)Select _All",  "<Control>A", self._select_all_cb, 0, ""),
      ("/Edit/_Rename",  "<Control>R", self._rename_cb, 0, ""),
      ("/Edit/_Delete",  "<Control>D", self._delete_cb, 0, ""),
      ("/Edit/(un)Randomi_ze",  "<Control>Z", self._random_order_cb, 0, ""),
      ("/Action",       None, None, 0, "<Branch>"),
      ("/Action/%s" % (ltext), "Return", self._activate, 0, ""),
      ("/Action/%s" % (itext), "<Shift>Return", self._activate_other, 0, ""),
      ]
    self.itemf.create_items(ks)
    self.menubar = self.itemf.get_widget("<main>")
    self.vbox.pack_start(self.menubar, 0, 0)

  def update_gui(self):
    ""
    #print "update_gui()"
    #self.clist.clear()
    self.list.clear()

    runqueue()

    #self.clist.freeze()
    for f in self.dir.files:
      title = f.basename
      if f.type == "/":
        title = title + "/"
      elif f.type == "*":
        title = title + "/"
      data = [title]
      #self.clist.append(data)

      self.list.append(data)

    self.listv.columns_autosize()
    if self.cfg.autoresize_window:
      runqueue()
      #wid,hgt = self.win.size_request()
      #print "%s,%s" % (wid, hgt)  # oops, same as requesting (1,1)
      self.win.resize(self.cfg.width, self.cfg.height)

    #self.clist.thaw()

  def go_dir(self, path="."):
    olddir = self.dir.path
    self._isearch = ""
    
    self.dir.loaddir(path)
    #print "go_dir(%s)" % (self.dir.path)
    self._cursor = 0
    self._selected = []

    if path == "..":
      import os.path
      dirs = os.path.split(olddir)
      olddir = dirs[-1]
      #print "Old dir: %s" % (olddir)

      
    self.update()

    if path == "..":
      self.select_by_name(olddir)

  def select_by_name(self, path):
    """TODO: make it work so it can match based on just the first letter
             for fast selection of files...
    """
    #print "select_by_name(%s)" % (path)
    old_cursor = self._cursor
    self._cursor = 0

    for i in xrange(len(self.dir.files)):
      compare = self.dir.files[i].basename
      #print "%s <--> %s" % (path, compare)
      c = compare[:len(path)]
      if path == compare:
	self._cursor = i
	break
      if path == c:
	self._cursor = i
	break
      if path < compare:
	self._cursor = i
	break
      if path > compare:
	self._cursor = i

    #print "Setting cursor to... %s" % (self._cursor)
    self.listv.set_cursor(self._cursor)

  def run(self):
    global num_dirs

    self.win.show_all()
    self.go_dir()
    #self.clist.grab_focus()
    self.listv.grab_focus()
    num_dirs = num_dirs + 1

  def update(self, *args):
    ""
    # Don't interrupt myself due to double-clicks
    if self._updating_now:
      return
    self._updating_now = 1  # cheapass mutex

    #self.status.set_text("Hello.")
    self._file_info()

    self.update_gui()
    self.win.set_title("pfm: %s" % (self.dir.path))

    self._updating_now = 0

  def _reload(self, *args):
    path = self.dir.files[self._cursor].basename
    self.go_dir()
    self.select_by_name(path)

  def _spawn_view(self, *args):
    d = dirwin(self.cfg, self.dir.path)
    d.run()

  def push_select(self, on):
    self._allow_select = [on] + self._allow_select
    #print "sel stack: %s" % (self._allow_select)

  def pop_select(self):
    del self._allow_select[0]
    #print "sel stack: %s" % (self._allow_select)

  def _selection_toggle(self, *args):
    #print "selection_toggle(%s)" % repr(args)
    if not self._allow_select[0]:
      #print "can't (de)select: blocked"
      return 0
    path = args[0]
    row = path[0]
    if self.dir.files[row] in self._selected:
      self._selected.remove(self.dir.files[row])
    else:
      self._selected.append(self.dir.files[row])
    #print "Selected:"
    #for f in self._selected:
    #  print "\t%s" % (f.basename)
    return 1

  def _select_all_cb(self, *args):
    self.push_select(1)
    if len(self._selected) > 0:
      self._unselect_all()
    else:
      self._select_all()
    self.pop_select()

  def _select_all(self, *args):
    self.push_select(1)
    self.selection.select_all()
    self.pop_select()

  def _unselect_all(self, *args):
    self.push_select(1)
    self.selection.unselect_all()
    self.pop_select()

  def _toggle_row(self, *args):
    self.push_select(1)
    path = args[1]
    row = path[0]
    if self.dir.files[row] in self._selected:
      self.selection.unselect_path(path)
    else:
      self.selection.select_path(path)
    self.pop_select()

  def _rename_cb(self, *args):
    self._rename()

  def _rename(self, *args):
    myfiles = self.get_selected_files()

    changed = 0
    for f in myfiles:
      fpath = "%s/%s" % (self.dir.path, f.basename)
      #print "Rename: %s" % (fpath)
      
      oldname = f.basename
      newname = f.basename
      (clicked, entered) = EntryWindow("Rename", oldname, ("", newname), transient=self.win)
      if clicked == "Ok":
	newname = entered
	if newname == "":
	  newname = oldname

      if oldname != newname:
	import os

	os.rename(fpath, "%s/%s" % (self.dir.path, newname))
	f.basename = newname
	changed = 1

    if changed: 
      #self.update()
      self.go_dir(".")

  def _delete_cb(self, *args):
    self._delete()

  def _delete(self, *args):
    myfiles = self.get_selected_files()
    buttons = ["Yes", "No"]
    if len(myfiles) > 1:
      buttons = ["Yes", "No", "Cancel"]

    changed = 0
    for f in myfiles:
      fpath = "%s/%s" % (self.dir.path, f.basename)
      #print "Delete: %s" % (fpath)
      
      clicked = ButtonChoiceWindow()("Delete?", "Delete %s?" % (f.basename),
                                   buttons, transient=self.win)
      if clicked == "Cancel":
	break

      elif clicked == "Yes":
	import os

	if f.type == "/":
	  os.rmdir(fpath)
	else:
	  os.unlink(fpath)
	changed = 1

    if changed: 
      #self.update()
      self.go_dir(".")

  def _random_order_cb(self, *args):
    if self.dir.sorttype == "random":
      self.dir.sorttype = "normal"
    else:
      self.dir.sorttype = "random"
    self.go_dir(".")
  
  def _listcb_move_cursor(self, *args):
    self.push_select(0)
    c, w = self.listv.get_cursor()
    #print "move cursor: %s" % repr(args)
    #print "Cursor: %s" % repr(c)
    self._cursor = c[0]
    self._last_was_isearch = 0
    self._isearch = ""
    self._file_info()

    if self.cfg.center_cursor:
      # doesn't always scroll right unless we draw first, then redraw
      runqueue()
      #print "scroll_to_cell((%s,), None, 1, 0.5, 0.5)" % (self._cursor)
      col = self.listv.get_column(0)
      self.listv.scroll_to_cell((self._cursor,), col, 1, 0.5, 1.0)
      runqueue()

    self.pop_select()
    
  def _listcb_keypress(self, cl, event, *args):
    if event.state & gtk.gdk.SHIFT_MASK:
      #print "Shift"
      actions = self._shift_actions
    else:
      actions = self._key_actions
    
    if actions.has_key(event.keyval):
      actions[event.keyval]()
    else:
      #print "Keypress: %s" % (event.string)
      punctuation = ".,-_[]{}()!'\";:~`+="
      if len(event.string) == 1:
	if (event.string >= "a"  and event.string <= "z")  or \
           (event.string >= "A"  and event.string <= "Z")  or \
	   (event.string >= "0"  and event.string <= "9")  or \
	   (event.string in punctuation):
	     if self._last_was_isearch:
	       self._isearch += event.string
	     else:
	       self._isearch = event.string
	     foo = self._isearch
	     self.select_by_name(self._isearch)
	     self._isearch = foo
	     self._last_was_isearch = 1
	     self._file_info()

  def _backspace(self, *args):
    if self._last_was_isearch:
      self._isearch = self._isearch[:-1]
    else:
      self._isearch = ""
    self.select_by_name(self._isearch)
    self._last_was_isearch = 1

  def _show_args(self, *args):
    print "Args: %s" % (repr(args))

  def _nop(self, *args):
    pass

  def _file_info(self):
    #print "file_info: %s" % (self._cursor)
    f = self.dir.files[self._cursor]
    if f.type == "/":
      text = "%s/ (DIR)" % (f.basename)
    elif f.type == "f":
      hsize = num_unit(f.size)
      text = "%s (%s)" % (f.basename, hsize)
    else:
      text = "%s/ (%s)" % (f.basename, f.type)
    if self._isearch:
      stext = self._isearch
      if len(stext) > 6:
	stext = self._isearch[:3] + ".." + self._isearch[-3:]
      text = stext + ": " + text
    self.status.set_text(text)

  def _up_dir(self, *args):
    #print "Up dir!"
    self.go_dir("..")

  def get_selected_files(self):
    #print "get_selected_files()"
    val = []
    current = self.dir.files[self._cursor]
    if len(self._selected) < 1:
      # return item under the cursor
      val = [current]
    else:
      # return all selected items, plus item under the cursor
      # oops, that's annoying.  just return selected items.
      #for f in self._selected:
      #  print f.basename
      #if current in self._selected:
      #  val.extend(self._selected)
      #else:
      #  val.append(current)
      #  val.extend(self._selected)
      val.extend(self._selected)

    return val

  def _activate_from_menu(self, *args):
    self._activate_auto()

  def _activate(self, *args):
    if self.cfg.activate_interactive:
      self._activate_interactive()
    else:
      self._activate_auto()

  def _activate_other(self, *args):
    if self.cfg.activate_interactive:
      self._activate_auto()
    else:
      self._activate_interactive()

  def _shell_escape(self, path):
    import string

    safe = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ" +\
           "1234567890-_,./"

    skip = 0
    i = 0
    flen = len(path)
    while i < flen:
      if path[i] not in safe:
	path = path[:i] + "\\" + path[i:]
	flen = flen + 1
	i = i + 1
	skip=1
      i = i + 1

    return path
  
  def _activate_auto(self, program=None):
    #print "Activate!"
    filestring = ""
    myfiles = self.get_selected_files()

    if myfiles[0].type == "/":
      self.go_dir(myfiles[0].basename)
      return

    if not program:
      program = self.get_action(myfiles[0])
    
    for f in myfiles:
      fpath = "%s/%s" % (self.dir.path, f.basename)
      fpath = self._shell_escape(fpath)
      filestring = "%s%s " % (filestring, fpath)
    cmd = "%s %s" % (program, filestring)
    #print cmd

    import os
    if os.fork() != 0:
      os.system(cmd)
      os._exit(1)

  def _activate_interactive(self, *args):
    #print "Activate (interactive)"
    filestring = ""
    myfiles = self.get_selected_files()

    if myfiles[0].type == "/":
      #fpath = self.dir.path + "/" + myfiles[0].basename
      #self._spawn_view(fpath)
      self.go_dir(myfiles[0].basename)
      return

    w = gtk.Window()
    w.set_modal(1)
    w.set_title("Choose Action")
    #w.set_default_size(200,200)
    w.connect("destroy", gtk.mainquit)

    vbox = gtk.VBox()
    w.add(vbox)

    l = gtk.Label("Choose Action:")
    vbox.pack_start(l, 0, 0)
    
    #sw = gtk.ScrolledWindow()
    #vbox.pack_start(sw, 1, 1)

    cl = gtk.CList(1)
    cl.set_column_auto_resize(0,1)
    #sw.add(cl)
    vbox.add(cl)
    cl.connect("select-row", self._activate_interactive_select)
    cl.connect("key-press-event", self._activate_interactive_keypress)
    cl.connect("button-release-event", self._activate_interactive_go)
    cl.set_selection_mode(gtk.SELECTION_BROWSE)

    self._activate_interactive_chosen = 0
    self._activate_interactive_cursor = 0
    actions = self.get_actions(myfiles[0])
    for a in actions:
      cl.append([a])
    cl.grab_focus()

    w.set_position(gtk.WIN_POS_MOUSE)
    w.show_all()
    gtk.mainloop()

    w.hide()
    w.unrealize()
    del w

    if self._activate_interactive_chosen:
      action=actions[self._activate_interactive_cursor]
      self._activate_auto(program=action)

  def _activate_interactive_keypress(self, cl, event, *args):
    if event.keyval in [gtk.gdk.keyval_from_name("Return"), 
                        gtk.gdk.keyval_from_name("space"), 
			gtk.gdk.keyval_from_name("Right")]:
      self._activate_interactive_chosen = 1
      gtk.mainquit()
    elif event.keyval in [gtk.gdk.keyval_from_name("Escape"), 
                          gtk.gdk.keyval_from_name("Left")]:
      self._activate_interactive_chosen = 0
      gtk.mainquit()
    
  def _activate_interactive_go(self, cl, event, *args):
    self._activate_interactive_chosen = 1
    gtk.mainquit()
      
  def _activate_interactive_select(self, cl, row, *args):
    self._activate_interactive_cursor = row
    
  def get_actions(self, f):
    import re

    actions = self.cfg.actions
    
    i = 0

    acts = []
    
    for i in xrange(len(actions)):
      pattern = actions[i][0]
      command = actions[i][1]
      fpath = "%s/%s" % (self.dir.path, f.basename)
      if re.search(pattern, fpath, re.I):
        acts.append(command)
        #print "Action: %s" % (command)
      
    return acts

  def get_action(self, f):
    acts = self.get_actions(f)
    return acts[0]

  def destroy(self, *args):
    global num_dirs

    self.win.hide()
    self.win.unrealize()
    gtk.mainquit()
    del self.win

    num_dirs = num_dirs - 1

def run(args):
  import pycfg
  cfg = pycfg.config(program_name)
  cfg.load()
  
  if len(args) > 0:
    for a in args:
      p = dirwin(cfg, a)
      p.run()
  else:
    p = dirwin(cfg)
    p.run()

  while num_dirs > 0:
    gtk.mainloop()

if __name__ == "__main__":
  import sys
  run(sys.argv[1:])
